//*********************************************************
// Copyright 2014 Intel Corporation All Rights Reserved.
// 
// The source code, information and material ("Material") contained herein is owned by Intel
// Corporation or its suppliers or licensors, and title to such Material remains with Intel 
// Corporation or its suppliers or licensors. The Material contains proprietary information 
// of Intel or its suppliers and licensors. The Material is protected by worldwide copyright 
// laws and treaty provisions. No part of the Material may be used, copied, reproduced, 
// modified, published, uploaded, posted, transmitted, distributed or disclosed in any way 
// without Intel's prior express written permission. No license under any patent, copyright 
// or other intellectual property rights in the Material is granted to or conferred upon 
// you, either expressly, by implication, inducement, estoppel or otherwise. Any license 
// under such intellectual property rights must be express and approved by Intel in writing.
// 
// Unless otherwise agreed by Intel in writing, you may not remove or alter this notice or 
// any other notice embedded in Materials by Intel or Intels suppliers or licensors in any 
// way.
//
// C_MainApp.h
// Definition of the C_MainApp class - high level code for a OpenCL application, with virtual functions
//
//*********************************************************



#pragma once


//  Main Application code - make modifications as necessary for your application


#include	"stdafx.h"
#include	"resource.h"





//	This is the "main application" base class - not intended to be instantiated more than once.
//		It helps organize access to the OpenCL environment.
//		A set of virtual functions, marked below, need to be defined in a derived class 
//			- see  class C_SampleOCLApp in SampleApp.h, and SampleApp.cpp
//		This class exists mainly to provide a clear separation of the simple application stages.
//		The "Run" function is used to run the application through its stages.
//		It creates and deletes the OpenCL environment, as defined by class C_OCL  (see C_OCL.h)

class C_MainApp
{
public:

		// Constructor with args
	C_MainApp( int  argc, _TCHAR*  argv[], cl_device_type  oclDevType )
	{
		Argc = argc;
		Argv = argv;

		// Construct OpenCL Environment
		OclEnv = new C_OCL( oclDevType );

	};

	~C_MainApp()
	{
		delete	OclEnv;
	};


	void	Run(  )
	{
		AppParseArgs();

		OclEnv->Start();	// run through start-up procedure for OpenCL environment

			// Application specific constructor - can assume OpenCL environment is created
		AppSetup();

		AppRun( );

		AppCleanup();

	};




protected:
	// Common Data
	int			Argc;
	_TCHAR **	Argv;
	C_OCL	*	OclEnv;



	// PURE VIRTUAL METHODS -  DEFINE THESE IN YOUR DERIVED CLASS
	virtual void	AppUsage() = 0;
	virtual void	AppParseArgs() = 0;
	virtual void	AppSetup() = 0;
	virtual void	AppCleanup() = 0;
	virtual void	AppRun() = 0;

};

