//*********************************************************
// Copyright 2014 Intel Corporation All Rights Reserved.
// 
// The source code, information and material ("Material") contained herein is owned by Intel
// Corporation or its suppliers or licensors, and title to such Material remains with Intel 
// Corporation or its suppliers or licensors. The Material contains proprietary information 
// of Intel or its suppliers and licensors. The Material is protected by worldwide copyright 
// laws and treaty provisions. No part of the Material may be used, copied, reproduced, 
// modified, published, uploaded, posted, transmitted, distributed or disclosed in any way 
// without Intel's prior express written permission. No license under any patent, copyright 
// or other intellectual property rights in the Material is granted to or conferred upon 
// you, either expressly, by implication, inducement, estoppel or otherwise. Any license 
// under such intellectual property rights must be express and approved by Intel in writing.
// 
// Unless otherwise agreed by Intel in writing, you may not remove or alter this notice or 
// any other notice embedded in Materials by Intel or Intels suppliers or licensors in any 
// way.
//
// C_OCL.h
// Definition of the C_OCL class
//
//*********************************************************


#pragma once

// OpenCL Environment  - reminder - this is sample code, NOT a wrapper - change it as you see fit!

#include	"stdio.h"
#include	<stdlib.h>

#include	<memory.h>
#include	<string.h>

#include	"CL/cl.h"
#include	"utils.h"

#define	OCL_ERROR	(-1)
#define	OCL_OK		(0)


#define		DEFAULT_PLATFORM_NAME		"Intel(R) OpenCL"	

#define	MAX_PLATFORM_NAMES	8

#define	GPU_CPU		(CL_DEVICE_TYPE_GPU | CL_DEVICE_TYPE_CPU)
#define	CPU_INDEX	0
#define	GPU_INDEX	1

#define		CL_ERR_CHECK_RETURN(_err_) 	{if( _err_ != CL_SUCCESS )	return	_err_;}



class C_OCL
{

public:

/////// Public Data ///////

		// To keep usage simpler, the following are public - but application should never set these
	cl_platform_id		CurPlatform;
	cl_device_id		CurDevice;
	cl_context			CurContext;				// Context for App to use
	cl_command_queue	CurQueue;



/////// Public Methods ///////

	C_OCL()
	{
		C_OCL( GPU_ONLY );		// default to GPU Only environment
	};

		// Create with ORing of device types to be supported - CPU and GPU currently supported (GPU_CPU)
	C_OCL( cl_device_type  envType );

	~C_OCL();


	cl_int		Start();	// Start the previously created OCL environment

	cl_int		SetCurrentDeviceType( cl_device_type proc );

	cl_int		GetDeviceInfo( cl_device_info  devInfo, void *  retData, size_t retSize );
	cl_int		GetDeviceInfo( cl_device_info  devInfo, void ** retData, size_t * retSize );

	cl_int		CheckExtension( char * chkstr, bool * b_supported );
	
	int			ReadSourceFile(const char *fileName, char * * retSrc );

	cl_program	CreateProgramFromFile( char * kernelFileName, int	* errRet );		// returns an index into Programs

	cl_kernel	CreateKernelFromProgram( cl_program prog,  char * kernelName,  int * errRet  );			// returns an index into Kernels

	void	ClearAllPrograms( );

	void	ClearAllKernels( );


		// Retrieve a kernel by index - application would need to record order of creation for indexing

	cl_kernel	GetKernel( int index )
	{
		return	Kernels[index];
	};



		// Print out all platform names

	void	PrintPlatformNames()
	{
		for( int i = 0; i < NumPlatforms; i++ )
		{
			printf("     %s   \n", PlatformNames[i] );
		}
	}



		// Select the current platform, by name

	void	SetCurPlatformName( char * platName )
	{
		if( SupportedPlatformName != NULL )
			free( SupportedPlatformName );

		SupportedPlatformName = _strdup( platName );
	};



private:

/////// Private Data ///////

		// Items the application should not need to directly access
	char			*		SupportedPlatformName; 
	cl_int					NumPlatforms;
	char			**		PlatformNames;			// to point to alloc'd array of pointers to strings
	cl_platform_id	*		PlatformIDs; 

	cl_device_type			AllowedDeviceTypes;
	cl_device_type			CurDeviceType;
	cl_device_id			DeviceIDs[2];			// GPU and Host device IDs - currently only 2 supported

	char			*		CurExtensionString;		// current device extension string

	cl_program		*		Programs;
	int						ProgCount;
	int						CurMaxProgs;

	cl_kernel		*		Kernels;
	int						KernCount;
	int						CurMaxKerns;


/////// Private Methods ///////
	
		// Manage Platforms & names, etc. the application should not need to directly access
	int		ReadAllPlatforms( );
	int		ClearAllPlatforms( );
	int		MatchPlatformName( char * platName );
	int		ReadExtensions( );


};

