//*********************************************************
// Copyright 2014 Intel Corporation All Rights Reserved.
// 
// The source code, information and material ("Material") contained herein is owned by Intel
// Corporation or its suppliers or licensors, and title to such Material remains with Intel 
// Corporation or its suppliers or licensors. The Material contains proprietary information 
// of Intel or its suppliers and licensors. The Material is protected by worldwide copyright 
// laws and treaty provisions. No part of the Material may be used, copied, reproduced, 
// modified, published, uploaded, posted, transmitted, distributed or disclosed in any way 
// without Intel's prior express written permission. No license under any patent, copyright 
// or other intellectual property rights in the Material is granted to or conferred upon 
// you, either expressly, by implication, inducement, estoppel or otherwise. Any license 
// under such intellectual property rights must be express and approved by Intel in writing.
// 
// Unless otherwise agreed by Intel in writing, you may not remove or alter this notice or 
// any other notice embedded in Materials by Intel or Intels suppliers or licensors in any 
// way.
//
// Performance.h
// Definition of useful macros for timing 
//
//*********************************************************


#pragma  once 

// Macro definition to measure performance
#include	<Windows.h>

	// USER CAN #define ENABLE_PERF_LAST to turn on performance measurement macros of last iteration measured
#ifdef ENABLE_PERF_LAST


#define	START_PERF( iters ) \
{  \
	LARGE_INTEGER __PerformanceFrequencyCountPerSec,  __PerformanceCountNDRangeStart, __PerformanceCountNDRangeStop;  \
	_tprintf(_T("%d iterations \n"), iters );	\
	float	millisec = 0.0f;	\
	QueryPerformanceFrequency( & __PerformanceFrequencyCountPerSec );	\
	for( int i = 0; i < iters; i++ )  \
	{   \
		QueryPerformanceCounter(&__PerformanceCountNDRangeStart);  




#define	END_PERF( ) \
		QueryPerformanceCounter(&__PerformanceCountNDRangeStop);  \
	}	\
	millisec = (float)(1000 * (__PerformanceCountNDRangeStop.QuadPart - __PerformanceCountNDRangeStart.QuadPart)) \
			/ (float) __PerformanceFrequencyCountPerSec.QuadPart; \
	_tprintf(_T("Last Measured Time:  %.2f  milliseconds \n"), millisec );	\
}


#else

#define	START_PERF_LAST( iters )
#define	END_PERF_LAST( )

#endif // ENABLE_PERF


	// USER CAN #define ENABLE_PERF_AVG to turn on performance measurement macros to report an average over some iterations
#ifdef ENABLE_PERF_AVG


#define	START_PERF( iters ) \
{  \
	int	 __iters = iters;	\
	_tprintf(_T("%d iterations \n"), iters );	\
	LARGE_INTEGER __PerformanceFrequencyCountPerSec,  __PerformanceCountNDRangeStart, __PerformanceCountNDRangeStop;  \
	float	millisec = 0.0f;	\
	QueryPerformanceFrequency( & __PerformanceFrequencyCountPerSec );	\
	QueryPerformanceCounter(&__PerformanceCountNDRangeStart);  \
	for( int i = 0; i < iters; i++ )  \
	{   




#define	END_PERF( ) \
	}	\
	QueryPerformanceCounter(&__PerformanceCountNDRangeStop);  \
	millisec =	( (float)(1000 * (__PerformanceCountNDRangeStop.QuadPart - __PerformanceCountNDRangeStart.QuadPart)) \
					/ (float) __PerformanceFrequencyCountPerSec.QuadPart )  \
					/ (float)__iters; \
	_tprintf(_T("Average Measured Time:  %.2f  milliseconds \n"), millisec );	\
}


#else

#define	START_PERF_AVG( iters )
#define	END_PERF_AVG( )

#endif // ENABLE_PERF

