//*********************************************************
// Copyright 2014 Intel Corporation All Rights Reserved.
// 
// The source code, information and material ("Material") contained herein is owned by Intel
// Corporation or its suppliers or licensors, and title to such Material remains with Intel 
// Corporation or its suppliers or licensors. The Material contains proprietary information 
// of Intel or its suppliers and licensors. The Material is protected by worldwide copyright 
// laws and treaty provisions. No part of the Material may be used, copied, reproduced, 
// modified, published, uploaded, posted, transmitted, distributed or disclosed in any way 
// without Intel's prior express written permission. No license under any patent, copyright 
// or other intellectual property rights in the Material is granted to or conferred upon 
// you, either expressly, by implication, inducement, estoppel or otherwise. Any license 
// under such intellectual property rights must be express and approved by Intel in writing.
// 
// Unless otherwise agreed by Intel in writing, you may not remove or alter this notice or 
// any other notice embedded in Materials by Intel or Intels suppliers or licensors in any 
// way.
//
// SampleApp.cpp
// Implementation of the virtual methods inherited from C_MainApp, plus some set-up for ToneMapping
//
//*********************************************************



// ToneMapping.cpp : Defines most application level code
//		Implements virtual methods for C_SampleOCLApp,  which use  C_OCL to help set up an OpenCL -based application
//		OpenCL kernel interface functions and other implementations are defined in separate files
//

#include "stdafx.h"




////////////////////////////////////////////////////////////////////////////////////////////////
// VIRTUAL METHODS INHERITED FROM C_MainApp - implement your OpenCL test application with these
////////////////////////////////////////////////////////////////////////////////////////////////

	//  Prints application usage - called by C_MainApp::Run()

void C_SampleOCLApp::AppUsage()
{
    _tprintf(_T("Usage: ToneMapping.exe [?|--h] [-c|-g]\n"));
    _tprintf(_T("  where, ? or --h prints this message\n"));
    _tprintf(_T("    -c run on CPU \n"));
    _tprintf(_T("    -g run on Processor Graphics - default\n"));
	_tprintf(_T("    -list  : display list of platform name strings\n"));
	_tprintf(_T("    -p \"platformName\"  :  supply a platform name (in quotes if it has spaces) to check for and use.\n")); 
	_tprintf(_T("    -i \"full Image filename\"  :  supply an image file name (in quotes if it has spaces) to process.\n")); 

	AppCleanup();

    exit(-1);
}



	// Parse command line arguments - called by C_MainApp::Run()
	//	No point making this more complex than necessary - want it to be easy to understand and modify.

void	C_SampleOCLApp::AppParseArgs()
{

	if( Argc < 2 )
	{
		_tprintf(_T("No command line arguments specified, using default values.\n"));

		OclEnv->SetCurrentDeviceType( CL_DEVICE_TYPE_GPU );		// default to use GPU
	
		return;
	}

	int argn = 1;
	while (argn < Argc)
	{
		if ((_tcscmp(Argv[argn], _T("--h")) == 0) || (_tcscmp(Argv[argn], _T("?")) == 0) || (_tcscmp(Argv[argn], _T("-?")) == 0))
		{
			AppUsage();
			argn++;
			if( argn == Argc )	exit(-1);
		}
		else if (_tcscmp(Argv[argn], _T("-list")) == 0)
		{
			_tprintf(_T("List of Platform Names: \n") );
			OclEnv->PrintPlatformNames();
			printf("\n");
			argn++;
			if( argn == Argc )	exit(-1);
		}
		else if (_tcscmp(Argv[argn], _T("-p")) == 0)
		{
			argn++;
				if( argn == Argc )	
				{
					_tprintf(_T("No platform name specified after \"-p\" \n") );
					exit(-1);
				}

			_tprintf(_T("Setting to use platform: <%s> \n\n"), Argv[argn] );
			OclEnv->SetCurPlatformName( CT2A( Argv[argn] ) );
			argn++;
		}
		else if (_tcscmp(Argv[argn], _T("-c")) == 0)
		{
			OclEnv->SetCurrentDeviceType( CL_DEVICE_TYPE_CPU );
			argn++;
		}
		else if (_tcscmp(Argv[argn], _T("-g")) == 0)
		{
			OclEnv->SetCurrentDeviceType( CL_DEVICE_TYPE_GPU );
			argn++;
		}
		else if (_tcscmp(Argv[argn], _T("-i")) == 0)
		{
			argn++;
				if( argn == Argc )	
				{
					_tprintf(_T("No input image file name specified after \"-i\" \n") );
					exit(-1);
				}

			_tprintf(_T("Setting input image: <%s> \n\n"), Argv[argn] );
			SetImageInName( Argv[argn] );
			argn++;
		}
		else
		{
			_tprintf(_T("Unknown command line argument:< %s >\n"), Argv[argn] );
			AppUsage();
			argn++;
			if( argn == Argc )	exit(-1);
		}
	}
};



	// Set up the OpenCL test application - called by C_MainApp::Run()
	//	This handles a lot of miscellaneous start-up stuff that usually clutters up "main()"

void	C_SampleOCLApp::AppSetup()
{
	int			err;
	cl_int		clErr;



		// Load bitmap image previously identified (user input or default)
	unsigned int	*	pInBitmap;
	float			*	inputArray;
	size_t 				imageWidth, imageHeight, imageSize;

	bool bOkBMP = LoadImageAsBMP ( CT2A( ImageFileIn ), &pInBitmap,  &imageWidth, &imageHeight, &imageSize );
		if ( ! bOkBMP )		exit(-1);	

	_tprintf(_T("\nInput Image %s has size %d x %d \n"), ImageFileIn, imageWidth, imageHeight );

		// ToneMapping kernel requires floating point input:
		// Convert input to floating point - required by ToneMap kernel
	size_t	pixelCnt = imageWidth * imageHeight;
	size_t	memSize;


		// Create Float4 image array from RGBA32bit bitmap
	ConvertUCHAR4Bitmap2FP4( pInBitmap, &inputArray, imageWidth, imageHeight, &memSize );

	_aligned_free( pInBitmap );			// no longer need integer version of original bitmap - converted to float4


	this->pInputImage = inputArray;
	this->ImageBufSize = memSize;
	this->ImageWidth = (int)imageWidth;
	this->ImageHeight = (int)imageHeight;



		// space for output from ToneMapping kernel
	this->pOclOutputImage = (float *) _aligned_malloc( memSize, 4096 );

		// space for output from Native ToneMapping code
	this->pNativeOutputImage = (float *) _aligned_malloc( memSize, 4096 );


	// Create buffer from bitmap image, and empty buffer to receive output
	this->cl_InBuf  = clCreateBuffer( OclEnv->CurContext,  CL_MEM_READ_WRITE | CL_MEM_USE_HOST_PTR, ImageBufSize, this->pInputImage, &clErr );
	this->cl_OutBuf = clCreateBuffer( OclEnv->CurContext,  CL_MEM_READ_WRITE | CL_MEM_USE_HOST_PTR, ImageBufSize, this->pOclOutputImage, &clErr );
		

		// Create the program and kernel for ToneMapping
	cl_program	prog  = OclEnv->CreateProgramFromFile( "ToneMapping.cl", &err );

	OclEnv->CreateKernelFromProgram( prog, "ToneMappingLine", &err );


	return;
}



	// Clean-up code for OpenCL test application - called at end by  C_MainApp::Run()
	//	Gets rid of stuff allocated for the application in  AppSetup()
void	C_SampleOCLApp::AppCleanup()
{
	if( pInputImage != NULL )
		_aligned_free( pInputImage );

	if( pOclOutputImage != NULL )
		_aligned_free( pOclOutputImage );

	if( pNativeOutputImage != NULL )
		_aligned_free( pNativeOutputImage );

}



	// Define virtual function - the core of the OpenCL test application 
	//		Called after set up by  C_MainApp::Run()  
	//	Deciding what goes in here instead of AppSetup or AppCleanup is a bit arbitrary
	//		but it's stuff considered tightly linked to actually running the ToneMapping code and getting results back

void	C_SampleOCLApp::AppRun()
{
	cl_int	clErr;

		// Set parameters in CHRData struct for ToneMap kernel, put in buffer to pass to kernel
		// init HDR parameters
    float kLow = -3.0f;
    float kHigh = 7.5f;
    float exposure = 3.0f;
    float gamma = 1.0f;
    float defog = 0.0f;

	SetHDRParams( &(this->ToneMapConfig), gamma,  defog,  kLow,  kHigh,  exposure );


		// Run the OpenCL version of ToneMapping
	clErr = RunOclToneMap(  );
		if( clErr != CL_SUCCESS )	_tprintf(_T("Error code %d \n"), clErr );


		// Pull bitmap from result buffer - map from OpenCL to CPU
		// Map the output back to CPU - assumes clFinish already called for kernel
	float * pfOutBuf = (float *)clEnqueueMapBuffer( OclEnv->CurQueue, this->cl_OutBuf, true, CL_MAP_READ, 0, this->ImageBufSize, 0, NULL, NULL, &clErr );


		// Convert output from float4 to uchar4 
	unsigned int	*	puiOutBuf;
	if( ! ConvertFP4Bitmap2UCHAR4( pfOutBuf, &puiOutBuf, this->ImageWidth, this->ImageHeight ) )
		_tprintf(_T("Failed to convert output buffer, float to packed uchar \n") );


		// release the floating point image bitmap back to OpenCL 
	clEnqueueUnmapMemObject( OclEnv->CurQueue, this->cl_OutBuf, pfOutBuf, 0, 0, 0 );


		// Write out OCL result bitmap
	SaveImageAsBMP ( puiOutBuf, this->ImageWidth, this->ImageHeight, "OclOutput.bmp" );

	_aligned_free( puiOutBuf );	// allocated in ConvertFP4Buffer2UCHAR4



		// Run the Native version of ToneMapping
	RunNativeToneMap(this->pInputImage, this->pNativeOutputImage, &(this->ToneMapConfig), this->ImageWidth,  this->ImageHeight );


		// Convert output from float4 to uchar4 
	if( ! ConvertFP4Bitmap2UCHAR4( this->pNativeOutputImage, &puiOutBuf, this->ImageWidth, this->ImageHeight ) )
		_tprintf(_T("Failed to convert output buffer, float to packed uchar \n") );

		// Write out OCL result bitmap
	SaveImageAsBMP ( puiOutBuf, this->ImageWidth, this->ImageHeight, "NativeOutput.bmp" );

	_aligned_free( puiOutBuf );	// allocated in ConvertFP4Buffer2UCHAR4


	return;
}




/////////////////////////////////////////////////////////////////////////////////////////
//	ToneMapping functions common to OpenCL and C++ implementations
/////////////////////////////////////////////////////////////////////////////////////////


	// Fill in HDRData struct 

void	SetHDRParams( CHDRData * pHDRData, float gamma, float defog, float kLow, float kHigh, float exposure )
{
	pHDRData->fGamma = gamma;
	pHDRData->fPowGamma = pow(2.0f, -3.5f*gamma);
	pHDRData->fDefog = defog;

	pHDRData->fPowKLow = pow( 2.0f, kLow );
	pHDRData->fPowKHigh = pow( 2.0f, kHigh );
	pHDRData->fPow35 = pow(2.0f, 3.5f);
	pHDRData->fPowExposure = pow( 2.0f, exposure +  2.47393f );

	//calculate FStops
	pHDRData->fFStops = resetFStopsParameter(pHDRData->fPowKLow, kHigh);
	//		_tprintf(_T("resetFStopsParameter result = %f\n"), pHDRData->fFStops);

	pHDRData->fFStopsInv = 1.0f/pHDRData->fFStops;
	
}



	// Calculate FStops value parameter from the arguments

float resetFStopsParameter( float powKLow, float kHigh )
{
    float curveBoxWidth = pow( 2.0f, kHigh ) - powKLow;
    float curveBoxHeight = pow( 2.0f, 3.5f )  - powKLow;

    // Initial boundary values
    float fFStopsLow = 0.0f;
    float fFStopsHigh = 100.0f;
    int iterations = 23; //interval bisection iterations

    // Interval bisection to find the final knee function fStops parameter 
    for ( int i = 0; i < iterations; i++ )
    {
        float fFStopsMiddle = ( fFStopsLow + fFStopsHigh ) * 0.5f;
        if ( ( curveBoxWidth * fFStopsMiddle + 1.0f ) < exp( curveBoxHeight * fFStopsMiddle ) )
        {
            fFStopsHigh = fFStopsMiddle;
        }
        else
        {
            fFStopsLow = fFStopsMiddle;
        }
    }

    return ( fFStopsLow + fFStopsHigh ) * 0.5f;
}



/////////////////////////////////////////////////////////////////////////////////////////
// Examples of using the C_OCL  GetDeviceInfo() function 
//		- not used in this simple application, but worth demonstrating.
//
//		Typical use would be getting OpenCL parameters to adapt to
//		different platforms' implementations of OpenCL.
/////////////////////////////////////////////////////////////////////////////////////////

void	exampleOfGetDeviceInfo( C_OCL *OclEnv )
{
	cl_int	clErr;



		//	GetDeviceInfo that malloc's space and returns a void pointer VARIABLE
		//		User responsible for freeing the space allocated (malloc-free)
		//		Also returns size via pointer
	char	*	versionString;		// note that all OpenCL strings are ASCII
	size_t		versionLen;

	clErr = OclEnv->GetDeviceInfo( CL_DEVICE_VERSION, (void **)&versionString, &versionLen );
		if( clErr != CL_SUCCESS )
		{
			_tprintf(_T("Error - unable to obtain version string. \n" ));
		}
		else
		{
			printf("GPU Device Version is [%s] \n", versionString );
			free( versionString );		// GetDeviceInfo malloc'd space for string - free it
		}



		// GetDeviceInfo that assumes app has created and passes pointer to appropriate space for return data
		//		Size is not returned 
	cl_ulong	localMem;

	clErr = OclEnv->GetDeviceInfo( CL_DEVICE_LOCAL_MEM_SIZE, &localMem, sizeof( localMem ) );
		if( clErr != CL_SUCCESS )
		{
			_tprintf(_T("Error - unable to obtain local memory size. \n" ) );
		}
		else
		{
			_tprintf(_T("Local Memory in bytes == %ld \n"), localMem );
		}

}

