//*********************************************************
// Copyright 2014 Intel Corporation All Rights Reserved.
// 
// The source code, information and material ("Material") contained herein is owned by Intel
// Corporation or its suppliers or licensors, and title to such Material remains with Intel 
// Corporation or its suppliers or licensors. The Material contains proprietary information 
// of Intel or its suppliers and licensors. The Material is protected by worldwide copyright 
// laws and treaty provisions. No part of the Material may be used, copied, reproduced, 
// modified, published, uploaded, posted, transmitted, distributed or disclosed in any way 
// without Intel's prior express written permission. No license under any patent, copyright 
// or other intellectual property rights in the Material is granted to or conferred upon 
// you, either expressly, by implication, inducement, estoppel or otherwise. Any license 
// under such intellectual property rights must be express and approved by Intel in writing.
// 
// Unless otherwise agreed by Intel in writing, you may not remove or alter this notice or 
// any other notice embedded in Materials by Intel or Intels suppliers or licensors in any 
// way.
//
// SampleApp.cpp
// Definition of the C_SampleOCLApp class, derived from C_MainApp
//		and some ToneMapping specific data/functions
//*********************************************************


#pragma once


#include	"stdafx.h"

// Application definitions specific to testing the ToneMapping kernel


	//	Parameters to control the ToneMapping algorithm
typedef struct 
{
    float fPowKLow;                             //fPowKLow = pow( 2.0f, kLow)
    float fPowKHigh;                            //fPowKHigh = pow( 2.0f, kHigh)
    float fPow35;                               //fPow35 = pow( 2.0f, 3.5f)
    float fFStops;                              //F stops
    float fFStopsInv;                           //Inverse fFStops value
    float fPowExposure;                         //fPowExposure = pow( 2.0f, exposure +  2.47393f )
    float fGamma;                               //Gamma correction parameter
    float fPowGamma;                            //Scale factor
    float fDefog;                               //Defog value

} CHDRData;




// ToneMap specific interface function prototypes

	// Function to set parameters for ToneMapping algorithm
void	SetHDRParams( CHDRData * HDRData, float gamma, float defog, float kLow, float kHigh, float exposure );

	//	A helper function called from SetHDRParams
float	resetFStopsParameter( float powKLow, float kHigh );


// Application may want to explicitly refer to one of multiple OpenCL Kernels
	//	C_OCL supports an array of multiple kernels
	//	There's only one kernel in this example - for simplicity, using a constant index to it
#define	TONE_MAP		0



	// This class groups application level values and methods.
	// Generally only a single instance of it will be instantiated.
	// It derives from C_MainApp  (See  C_MainApp.h) and will implement the virtual functions defined there.

class C_SampleOCLApp: public C_MainApp
{
public:

	C_SampleOCLApp(int  argc, _TCHAR*  argv[], cl_device_type  oclDevType) : C_MainApp(argc, argv, oclDevType )
	{
		ImageFileIn = (_TCHAR *)NULL;

		SetImageInName( _T("InputImage.bmp") );
		pInputImage = NULL;
		pOclOutputImage = NULL;
		pNativeOutputImage = NULL;
	}


	// Kernel specific defined application specific data
	C_SampleOCLApp::~C_SampleOCLApp()
	{
	}


private:
	_TCHAR		*	ImageFileIn;
	float		*	pInputImage;
	float		*	pOclOutputImage;
	float		*	pNativeOutputImage;

	size_t			ImageBufSize;
	int				ImageWidth, ImageHeight;

	cl_mem			cl_InBuf;
	cl_mem			cl_OutBuf;

	CHDRData		ToneMapConfig;


		// A simple helper method to set the single input image file name by copying the string
	void	SetImageInName( _TCHAR * fName )
	{
		if( ImageFileIn != NULL )
			free( ImageFileIn );

		ImageFileIn = _tcsdup( fName );
	};



		//  Prototypes of interface functions for OpenCL ToneMapping kernel, defined in ToneMap_OCL.cpp
	cl_int		RunOclToneMap(  );
	cl_int		ToneMap( cl_mem	inBuf, cl_mem	outBuf, int width, int height, int stride, cl_mem pParams  );
	
	
		// Prototypes of interface functions for  C++ implementation of ToneMap defined in ToneMap_Native.cpp
	void		RunNativeToneMap( float* inputArray, float* outputArray, CHDRData *pData, int arrayWidth, int arrayHeight );
	void		EvaluateRaw( float* inputArray, float* outputArray, CHDRData *pData, int arrayWidth, int arrayHeight);


		// Virtual functions from C_MainApp - defined in SampleApp.cpp
	void	AppUsage();
	void	AppParseArgs();
	void	AppSetup();
	void	AppRun();
	void	AppCleanup();



};
