//*********************************************************
// Copyright 2009-2014 Intel Corporation All Rights Reserved.
// 
// The source code, information and material ("Material") contained herein is owned by Intel
// Corporation or its suppliers or licensors, and title to such Material remains with Intel 
// Corporation or its suppliers or licensors. The Material contains proprietary information 
// of Intel or its suppliers and licensors. The Material is protected by worldwide copyright 
// laws and treaty provisions. No part of the Material may be used, copied, reproduced, 
// modified, published, uploaded, posted, transmitted, distributed or disclosed in any way 
// without Intel's prior express written permission. No license under any patent, copyright 
// or other intellectual property rights in the Material is granted to or conferred upon 
// you, either expressly, by implication, inducement, estoppel or otherwise. Any license 
// under such intellectual property rights must be express and approved by Intel in writing.
// 
// Unless otherwise agreed by Intel in writing, you may not remove or alter this notice or 
// any other notice embedded in Materials by Intel or Intels suppliers or licensors in any 
// way.
//
// ToneMap_Native.cpp
// Implementation of the Native (CPU) version of the ToneMapping algorithm
//  Derived from  Intel(R) SDK for OpenCL* Applications ToneMapping sample code
//
//*********************************************************

#include "stdafx.h"

//	Enable/disable performance measurement around the kernel enqueue
//#define		ENABLE_PERF_LAST	1
#define		ENABLE_PERF_AVG	1
#include	"Performance.h"


//	Runs the C++ implementation of ToneMapping, takes performance measurements

void	C_SampleOCLApp::RunNativeToneMap(float* inputArray, float* outputArray, CHDRData *pData, int arrayWidth, int arrayHeight)
{

#if ENABLE_PERF_LAST || ENABLE_PERF_AVG
	_tprintf(_T("\n\nMEASURING PERFORMANCE OF Native C++ ToneMapping \n") );
#endif // ENABLE_PERF_LAST

	START_PERF(10 );

		// execute the native implementation
	EvaluateRaw( inputArray, outputArray, pData, arrayWidth,  arrayHeight );

	END_PERF( );


}


//	A C++ implementation of ToneMapping

void C_SampleOCLApp::EvaluateRaw(float* inputArray, float* outputArray, CHDRData *pData, int arrayWidth, int arrayHeight)
{
    __m128 fPowKLow = _mm_set_ps1( pData->fPowKLow);
    __m128 fFStops = _mm_set_ps1( pData->fFStops );
    __m128 fFStopsInv = _mm_set_ps1(pData->fFStopsInv);
    __m128 fPowExposure = _mm_set_ps1(pData->fPowExposure);
    __m128 fDefog = _mm_set_ps1(pData->fDefog);
    __m128 fGamma = _mm_set_ps1(pData->fGamma);
    __m128 fPowGamma = _mm_set_ps1(pData->fPowGamma);

    // and define method constants.
    __m128 fOne = _mm_set_ps1( 1.0f );
    __m128 fZero = _mm_set_ps1( 0.0f );
    __m128 fSaturate = _mm_set_ps1( 255.f );

	for( int iRow = 0; iRow < arrayHeight; iRow++ )
	{
		for (int iCol = 0; iCol < arrayWidth; iCol++)
		{
			__m128 fColor = _mm_setzero_ps( );
			fColor = _mm_load_ps((float*)((char*)inputArray+iRow*arrayWidth*4*sizeof(float)+ iCol*4*sizeof(float)));

			//Defog 
			fColor = _mm_sub_ps(fColor, fDefog);
			fColor = _mm_max_ps(fZero, fColor);

			// Multiply color by pow( 2.0f, exposure +  2.47393f )
			fColor = _mm_mul_ps( fColor, fPowExposure ); 

			// Apply a knee function.
			__m128 fCmpFlag = _mm_cmpge_ps( fColor, fPowKLow );

			if(( _mm_movemask_ps( fCmpFlag ) & 7) != 0 ) 
			{
				__m128 fTmpPixel = _mm_sub_ps( fColor, fPowKLow );
				fTmpPixel = _mm_mul_ps( fTmpPixel, fFStops );
				fTmpPixel = _mm_add_ps( fTmpPixel, fOne );

				//fTmpPixel = logf( fTmpPixel);
				((float*)&fTmpPixel)[0] = logf(((float*)&fTmpPixel)[0]); 
				((float*)&fTmpPixel)[1] = logf(((float*)&fTmpPixel)[1]);
				((float*)&fTmpPixel)[2] = logf(((float*)&fTmpPixel)[2]);
				((float*)&fTmpPixel)[3] = logf(((float*)&fTmpPixel)[3]);    

				fTmpPixel = _mm_mul_ps( fTmpPixel, fFStopsInv );
				fColor = _mm_add_ps( fTmpPixel, fPowKLow );
			}

			//Gamma correction
			//fColor = pow(fColor, fGamma);
			((float*)&fColor)[0] = pow(((float*)&fColor)[0], ((float*)&fGamma)[0]); 
			((float*)&fColor)[1] = pow(((float*)&fColor)[1], ((float*)&fGamma)[1]);
			((float*)&fColor)[2] = pow(((float*)&fColor)[2], ((float*)&fGamma)[2]);
			((float*)&fColor)[3] = pow(((float*)&fColor)[3], ((float*)&fGamma)[3]);    

			// Scale the values
			fColor = _mm_mul_ps( fColor, fSaturate );
			fColor = _mm_mul_ps( fColor, fPowGamma );

			//Saturate
			fColor = _mm_max_ps( fColor, _mm_setzero_ps( ));
			fColor = _mm_min_ps( fColor, fSaturate );

			//Store result pixel 
			_mm_store_ps((float*)((char*)outputArray+iRow*arrayWidth*4*sizeof(float)+ iCol*4*sizeof(float)), fColor);
		}
	}
}


