//*********************************************************
// Copyright 2009-2014 Intel Corporation All Rights Reserved.
// 
// The source code, information and material ("Material") contained herein is owned by Intel
// Corporation or its suppliers or licensors, and title to such Material remains with Intel 
// Corporation or its suppliers or licensors. The Material contains proprietary information 
// of Intel or its suppliers and licensors. The Material is protected by worldwide copyright 
// laws and treaty provisions. No part of the Material may be used, copied, reproduced, 
// modified, published, uploaded, posted, transmitted, distributed or disclosed in any way 
// without Intel's prior express written permission. No license under any patent, copyright 
// or other intellectual property rights in the Material is granted to or conferred upon 
// you, either expressly, by implication, inducement, estoppel or otherwise. Any license 
// under such intellectual property rights must be express and approved by Intel in writing.
// 
// Unless otherwise agreed by Intel in writing, you may not remove or alter this notice or 
// any other notice embedded in Materials by Intel or Intels suppliers or licensors in any 
// way.
//
// ToneMap_OCL.cpp
// Implementation of the OpenCL v1.2 version of the host code to invoke ToneMapping kernel
//  Derived from  Intel(R) SDK for OpenCL* Applications ToneMapping sample code
//
//*********************************************************




//	Place Kernel Specific Code in this type of file
//	Split into two main functions 
//		One that generically encapsulates the OpenCL kernel, without application specific information
//		One that is application specific - access app data to set up and call the generic kernel interface function


#include	"stdafx.h"



	//	Enable/disable performance measurement around the kernel enqueue

//#define		ENABLE_PERF_LAST		1
#define		ENABLE_PERF_AVG		1

#include	"Performance.h"



//	Interface from the host application 
//	Calls ToneMap() to execute the kernel

cl_int	C_SampleOCLApp::RunOclToneMap( ) 
{
	cl_int	clErr;


		// Create a buffer object with copy of CHRDATA * pParams 
	cl_mem	params = clCreateBuffer( OclEnv->CurContext, (CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR), sizeof(CHDRData), &(this->ToneMapConfig), &clErr );
 		CL_ERR_CHECK_RETURN(clErr);

#if ENABLE_PERF_LAST || ENABLE_PERF_AVG
	_tprintf(_T("\n\nMEASURING PERFORMANCE OF OpenCL ToneMapping \n") );
#endif // ENABLE_PERF_x

	clErr = ToneMap( this->cl_InBuf, this->cl_OutBuf,  this->ImageWidth,  this->ImageHeight,  this->ImageWidth, params  );
 		CL_ERR_CHECK_RETURN(clErr);


	clFinish( OclEnv->CurQueue );

	clReleaseMemObject( params );


	return CL_SUCCESS;

}



//	Interface to execute the OpenCL ToneMapping kernel 
//		Many applications create wrapper code to avoid re-implementing this code pattern for each kernel.
//		However, simply copying and modifying this code for each kernel supported, can make it
//		 easy to specialize for each kernel, and easy to understand/debug/maintain.

cl_int		C_SampleOCLApp::ToneMap( cl_mem	inBuf, cl_mem	outBuf, int width, int height, int stride, cl_mem pParams  )
{
	cl_kernel			kern	= OclEnv->GetKernel(TONE_MAP);		// only one kernel in this example!
	cl_context			context = OclEnv->CurContext;
	cl_device_id		device	= OclEnv->CurDevice;
	cl_command_queue	queue	= OclEnv->CurQueue;
	cl_int				err		= CL_SUCCESS;


		// SET KERNEL PARAMETERS TO BE PASSED
	int	argI = 0;
	err = CL_SUCCESS;
	err |= clSetKernelArg( kern, argI++, sizeof(cl_mem), &inBuf );
	err |= clSetKernelArg( kern, argI++, sizeof(cl_mem), &outBuf );
	err |= clSetKernelArg( kern, argI++, sizeof(cl_mem), &pParams );
	err |= clSetKernelArg( kern, argI++, sizeof(width),  &width );
	err |= clSetKernelArg( kern, argI++, sizeof(height), &height );
 		CL_ERR_CHECK_RETURN(err);

		
		// Global workspace for this kernel is just 1 dimension == width - kernel processes down columns of pixels
    size_t globalWorkSize[2];
	globalWorkSize[0] =	width;
    //	globalWorkSize[1] =	1;

		// This kernel can use null pointer for local work group size, letting runtime choose best size
    //	size_t localWorkSize[2];
	//	localWorkSize[0] =	16;
	//	localWorkSize[1] =	1;


	START_PERF(100 );

		// execute kernel - Allowing local WG size to default, set by runtime
    err = clEnqueueNDRangeKernel(queue, kern, 1, NULL, globalWorkSize, NULL, 0, NULL, NULL);
 		CL_ERR_CHECK_RETURN(err);

    err = clFinish(queue);		// a "real" application would probably not call this - needed for performance measurement.

	END_PERF( );


	return 0;
}





