//*********************************************************
// Copyright 2009-2014 Intel Corporation All Rights Reserved.
// 
// The source code, information and material ("Material") contained herein is owned by Intel
// Corporation or its suppliers or licensors, and title to such Material remains with Intel 
// Corporation or its suppliers or licensors. The Material contains proprietary information 
// of Intel or its suppliers and licensors. The Material is protected by worldwide copyright 
// laws and treaty provisions. No part of the Material may be used, copied, reproduced, 
// modified, published, uploaded, posted, transmitted, distributed or disclosed in any way 
// without Intel's prior express written permission. No license under any patent, copyright 
// or other intellectual property rights in the Material is granted to or conferred upon 
// you, either expressly, by implication, inducement, estoppel or otherwise. Any license 
// under such intellectual property rights must be express and approved by Intel in writing.
// 
// Copyright (c) Microsoft. All rights reserved.
// 
// THIS CODE IS PROVIDED *AS IS* WITHOUT WARRANTY OF
// ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING ANY
// IMPLIED WARRANTIES OF FITNESS FOR A PARTICULAR
// PURPOSE, MERCHANTABILITY, OR NON-INFRINGEMENT.
// 
// Unless otherwise agreed by Intel in writing, you may not remove or alter this notice or 
// any other notice embedded in Materials by Intel or Intels suppliers or licensors in any 
// way.
//
// utils.cpp
// Implementation of a few utility functions
//	SaveImageAsBMP function is from  Intel(R) SDK for OpenCL* Applications ToneMapping sample code
//
//*********************************************************

#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>
#include <memory.h>
#include <windows.h>
//#include "CL\cl.h"
//#include "CL\cl_ext.h"
#include "utils.h"
#include <assert.h>


//we want to use POSIX functions
#pragma warning( push )
#pragma warning( disable : 4996 )




#define	RGBA_ORDER 0

bool LoadImageAsBMP ( const char* fileName, unsigned int ** pBuf, size_t *pWidth, size_t *pHeight, size_t	*pMemSize )
{
    FILE* stream;

	stream = fopen( fileName, "rb" );

	if( stream == NULL )
	{
		printf("Unable to open file %s !\n", fileName );
		return false;
	}

    BITMAPFILEHEADER fileHeader;
    BITMAPINFOHEADER infoHeader;
    size_t iMemSize = 0;

    if( sizeof(BITMAPFILEHEADER) != fread( &fileHeader, 1, sizeof(BITMAPFILEHEADER), stream ) ) {
        // can't read BITMAP FILE HEADER
       printf("Unable to read header from file %s invalid!\n", fileName );
       goto ErrExit;
    }

    if( sizeof(BITMAPINFOHEADER) != fread( &infoHeader, 1, sizeof(BITMAPINFOHEADER), stream ) ) {
        // can't read BITMAP INFO HEADER
        goto ErrExit;
    }

	unsigned int		width = infoHeader.biWidth;
	unsigned int		height = infoHeader.biHeight;

    if(width<=0 || height<=0 || width > 1000000 || height > 1000000)
    {
        printf("HOST: width or height values are invalid!\n");
        return 0;
    }

	*pWidth = width;
	*pHeight = height;
	
    //! The image size in memory (bytes).
    iMemSize = (size_t)(width*height*sizeof(unsigned int)); 
	iMemSize = (iMemSize + 63) & ~0x00ff;	// intel OpenCL requires size to be multiple of 64 byte cache line
	*pMemSize = iMemSize;


    //! Allocate memory.
    unsigned int *  pTmp = (unsigned int *)_aligned_malloc(iMemSize, 4096 );
    if(! pTmp)
    {
        printf("Failed to allocate memory for input HDR image!\n");
        goto ErrExit;
    }


    char* ppix = (char*)pTmp;


    size_t rowLength = width * 3;

	unsigned char * buffer = (unsigned char*)malloc( rowLength + 4);

    unsigned int x, y;

    for (y=0; y<height; y++)
    {
		ppix = (char *)(pTmp + (height-y-1)*width);
		if( rowLength != fread(buffer, 1, rowLength, stream )) 
		{
			goto ErrExit;
        }
		for( x = 0; x < 3*width; x+=3 )
		{
#if RGBA_ORDER	//  Reversing to RGBA order for OpenCL processing
			*ppix++ = buffer[x+2];	
			*ppix++ = buffer[x+1];
			*ppix++ = buffer[x+0];
#else	//  BGRA order - default
			*ppix++ = buffer[x+0];	
			*ppix++ = buffer[x+1];
			*ppix++ = buffer[x+2];
#endif
			*ppix++ = (BYTE)0x0ff;	// no alpha channel - set to 0xff
        }
    }


	// set return pointer with allocated and filled buffer
	*pBuf = pTmp;

	// DEBUG LINE = UNCOMMENT TO GET COPY OF INPUT AS OUTPUT
	//SaveImageAsBMP( (unsigned int *)pTmp, width, height, "CopyInput.bmp");

    fclose( stream );
    return true;


ErrExit:
    fclose( stream );
    return false;
}




bool SaveImageAsBMP ( unsigned int* ptr, int width, int height, const char* fileName)
{
    FILE* stream = NULL;
    int* ppix = (int*)ptr;
    printf("Save Image: %s \n", fileName);
    stream = fopen( fileName, "wb" );

	if( NULL == stream )
		return false;

    BITMAPFILEHEADER fileHeader;
    BITMAPINFOHEADER infoHeader;

    int alignSize  = width * 4;
    alignSize ^= 0x03;
    alignSize ++;
    alignSize &= 0x03;

    int rowLength = width * 4 + alignSize;

    fileHeader.bfReserved1  = 0x0000;
    fileHeader.bfReserved2  = 0x0000;

    infoHeader.biSize          = sizeof(BITMAPINFOHEADER);
    infoHeader.biWidth         = width;
    infoHeader.biHeight        = height;
    infoHeader.biPlanes        = 1;
    infoHeader.biBitCount      = 24;
    infoHeader.biCompression   = BI_RGB;
    infoHeader.biSizeImage     = rowLength * height;
    infoHeader.biXPelsPerMeter = 0;
    infoHeader.biYPelsPerMeter = 0;
    infoHeader.biClrUsed       = 0; // max available
    infoHeader.biClrImportant  = 0; // !!!
    fileHeader.bfType       = 0x4D42;
    fileHeader.bfSize       = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER) + rowLength * height;
    fileHeader.bfOffBits    = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER);

    if( sizeof(BITMAPFILEHEADER) != fwrite( &fileHeader, 1, sizeof(BITMAPFILEHEADER), stream ) ) {
        // cann't write BITMAPFILEHEADER
        goto ErrExit;
    }

    if( sizeof(BITMAPINFOHEADER) != fwrite( &infoHeader, 1, sizeof(BITMAPINFOHEADER), stream ) ) {
        // cann't write BITMAPINFOHEADER
        goto ErrExit;
    }

    unsigned char buffer[4];
    int x, y;

    memset( buffer, 0x00, 4 );

    for (y=0; y<height; y++)
    {
		ppix = (int *)(ptr + (height-y-1)*width);

        for (x=0; x<width; x++, ppix++)
        {
            if( 3 != fwrite(ppix, 1, 3, stream))	// inefficient? - but simple.  Writing BGR order
			{
                goto ErrExit;
            }
        }
        fwrite( buffer, 1, alignSize, stream );
    }

    fclose( stream );
    return true;
ErrExit:
    fclose( stream );
    return false;
}



//	Convert floating point RGBA to packed unsigned char RGBA
//		Assumes buffer has stride=width
//		Returns allocated space 
bool		ConvertFP4Bitmap2UCHAR4( float * inBuf, unsigned int * * ppOutBuf, size_t width, size_t height )
{
	size_t				pixCnt = width * height;
	unsigned int	*	pOutBuf;

    pOutBuf = (unsigned int*)_aligned_malloc(pixCnt * sizeof(unsigned int), 4096);
		if( ! pOutBuf )
		{
			printf("Failed to allocate memory for packed uchar4 image!\n");
			return false;
		}

	float			*	pfInBuf = inBuf;
	unsigned int	*	pTmpOut = pOutBuf;
    unsigned int		uiTmp0, uiTmp1, uiTmp2, uiTmp3;
	unsigned int		uiResult;

	for( size_t	pix = 0; pix < pixCnt; pix++ )
	{
		uiTmp0 = (unsigned int)(*pfInBuf);
			pfInBuf++;
		uiTmp1 = (unsigned int)(*pfInBuf);
			pfInBuf++;
		uiTmp2 = (unsigned int)(*pfInBuf);
			pfInBuf++;
		uiTmp3 = (unsigned int)(*pfInBuf);
			pfInBuf++;

	    uiResult  = 0x000000FF & uiTmp0;
        uiResult |= 0x0000FF00 & (uiTmp1 << 8);
        uiResult |= 0x00FF0000 & (uiTmp2 << 16);
        uiResult |= 0xFF000000 & (uiTmp3 << 24);

		*pTmpOut = uiResult;

		pTmpOut++;
	}

	*ppOutBuf = pOutBuf;

	return	true;

}




//	Convert packed unsigned char RGBA to floating point RGBA  
//		Assumes buffer has stride=width
//		Returns allocated space - caller must use _aligned_free()
bool		ConvertUCHAR4Bitmap2FP4( unsigned int  * inBuf, float * * ppOutBuf, size_t width, size_t height, size_t *pMemSize )
{
	size_t			pixCnt = width * height;
	float		*	pOutBuf;

		// Creating a buffer of four floats per color pixel, rounding size up to next multiple of 64 for Intel OpenCL
	size_t			memSize = ((pixCnt * 4 * sizeof(float)) + 63) & ~0x003f;	

    pOutBuf = (float *)_aligned_malloc(memSize, 4096);
		if( ! pOutBuf )	{
			printf("Failed to allocate memory for packed float4 image!\n");
			return false;
		}


	unsigned int	*	puiInBuf = inBuf;
	float			*	pTmpOut = pOutBuf;

	for( size_t	pix = 0; pix < pixCnt; pix++ )
	{
		*pTmpOut = (float)(0x000000FF & (*puiInBuf)) / 255.0f;
			pTmpOut++;
		*pTmpOut = (float)((0x0000FF00 & (*puiInBuf)) >> 8) / 255.0f;
			pTmpOut++;
		*pTmpOut = (float)((0x00FF0000 & (*puiInBuf)) >> 16) / 255.0f;
			pTmpOut++;
		*pTmpOut = (float)((0xFF000000 & (*puiInBuf)) >> 24) / 255.0f;
			pTmpOut++;

		puiInBuf++;

	}

	*ppOutBuf = pOutBuf;
	*pMemSize = memSize;

	return	true;

}


#pragma warning( pop )
